/* Sample handling:
 * - loading.
 * - playing.
 * - playing rate changing.
 * - sample mem freeing.
 */

#include <stdio.h>
#include <unistd.h>
#include <stat.h>
#include <time.h>
#include <signal.h>
#include <memory.h>
#include "action.h"

/* from bitmaps.c */
extern void ShowRate(int rate);
extern void ShowSample(char *name, int secs);
extern void ReverseButton(int button);
extern void NormalButton(int button);

/* playback.c */
extern playback(unsigned char *sample, long lenght, int rate);

char *FileNames[FILENAMES];
static int SampleIdx = 0;		/* index in filenames */

static unsigned char *Sample = 0;	/* sample data */
static long SampleLenght;		/* in bytes */
static int SampleSecs;			/* in seconds */
static int SampleRate = 8;		/* sample playback rate */
static int SamplePlaying = 0;
static int SampleSign = NORMAL;

#define SAMPLESECS ((SampleLenght / 1024) / SampleRate + 1)

/* free sample mem */
void DoClean(void)
{
  sigset_t sigmask, osigmask;
  
  if(Sample)
  {
    /* Block SIGALRM until we'll ready to synchronize with it. (POSIX) */
    sigemptyset(&sigmask);
    sigaddset(&sigmask, SIGALRM);
    sigprocmask(SIG_BLOCK, &sigmask, &osigmask);

    /* Wait until playing ends before freeing the sample buffer. */
    if(SamplePlaying)
      sigsuspend(&osigmask);
      
    /* Obviously we're now synchronized */
    sigprocmask(SIG_SETMASK, &osigmask, 0);
    free(Sample);
    Sample = 0;
    SampleLenght = 0;
  }
}

/* timeout handler */
void SampleDone(int sig) { SamplePlaying = 0; }

/* Load a sample */
int LoadFile(char *name)
{
  FILE *fp;
  struct stat fst;

  if((fp = fopen(name, "rb")))
  {
    DoClean();			/* remove previous sample */
    stat(name, &fst);
    if((Sample = (unsigned char *)malloc(fst.st_size)))
    {
      SampleLenght = fst.st_size;
      fread(Sample, 1, SampleLenght, fp);
      SampleSecs = SAMPLESECS;
      SampleSign = NORMAL;
    }
    else
    {
      fprintf(stderr, "Not enough memory: %ld\n", fst.st_size);
      return 0;
    }
    fclose(fp);
  }
  else
  {
    fprintf(stderr, "Unable to open %s\n", name);
    return 0;
  }
  return 1;
}

/* Draw window content (buttons etc.) */
void Redraw(void)
{
  int i;
  ShowSample(FileNames[SampleIdx], SampleSecs);
  for(i = 0; i < BITMAPS; i++)
  {
    if(i == SIGN && SampleSign == REVERSE)
      ReverseButton(i);
    else
      NormalButton(i);
  }
  ShowRate(SampleRate);
}

/* signed/unsigned sample conversion */
void InvertSample(unsigned char *addr, long lenght)
{
  for(; lenght > 0; lenght--)
    *(addr++) ^= 0x80;
}

/* user actions */
int DoAction(char action)
{
  int success = 1;

  switch(action)
  {
    case OpPREV:
      ReverseButton(PREV);
      if(SampleIdx > 0)
      {
        success = LoadFile(FileNames[--SampleIdx]);
        Redraw();
      }
      NormalButton(PREV);
      break;

    case OpNEXT:
      ReverseButton(NEXT);
      if(FileNames[SampleIdx+1] != 0)
      {
        success = LoadFile(FileNames[++SampleIdx]);
        Redraw();
      }
      NormalButton(NEXT);
      break;

    case OpPLAY:
      ReverseButton(PLAY);
      if(Sample)
      {
        playback(Sample, SampleLenght, SampleRate);
        /* set alarm for sample play end */
	signal(SIGALRM, SampleDone);		/* set handler */
	SamplePlaying = 1;
	alarm(SampleSecs);			/* set signal  */
      }
      NormalButton(PLAY);
      break;

    case OpSTOP:
      ReverseButton(STOP);
      return 0;
      NormalButton(STOP);
      break;

    case OpDOWN:
      ReverseButton(DOWN);
      if(SampleRate > 2)
      {
        SampleRate--;
	SampleSecs = SAMPLESECS;
        Redraw();
      }
      NormalButton(DOWN);
      break;

    case OpUPPE:
      ReverseButton(UPPE);
      if(SampleRate < 14)
      {
        SampleRate++;
	SampleSecs = SAMPLESECS;
        Redraw();
      }
      NormalButton(UPPE);
      break;

    case OpSIGN:
      InvertSample(Sample, SampleLenght);
      if(SampleSign == NORMAL)
      {
        SampleSign = REVERSE;
	ReverseButton(SIGN);
      }
      else
      {
        SampleSign = NORMAL;
	NormalButton(SIGN);
      }
      break;
  }
  return success;
}
